from __future__ import annotations

from typing import Any, Self, MutableMapping, Sequence, get_origin, get_args, Collection, Mapping
import collections
import inspect
from types import NoneType
from abc import ABC, abstractmethod

#

# Used for annotations
JSONArray = list["JSONType"]
JSONObject = dict[str, "JSONType"]
JSONValue = None | bool | int | float | str
JSONContainer = JSONObject | JSONArray
JSONType = JSONContainer | JSONValue

FieldPath = str | int | tuple[str | int, ...]


class SprelfJSONError(Exception):

    def __init__(self, *args):
        super().__init__(*args)


#


class _JSONObjectLike(type):
    ALLOWED = (dict, MutableMapping, Mapping, collections.abc.MutableMapping, collections.abc.Mapping)

    def __instancecheck__(cls, obj: Any) -> bool:
        return obj is not None and isinstance(obj, cls.ALLOWED) and \
            all(isinstance(v, JSONLike) for v in obj.values()) and all(isinstance(k, str) for k in obj.keys())

    def __subclasscheck__(cls, t: type) -> bool:
        if t == JSONObjectLike:
            return True
        origin = get_origin(t)
        args = get_args(t)
        return ((inspect.isclass(origin) and issubclass(origin, cls.ALLOWED)) or origin in cls.ALLOWED) and \
                len(args) == 2 and args[0] == str and issubclass(args[1], JSONLike)


class JSONObjectLike(metaclass=_JSONObjectLike):
    ...


class _JSONValueLike(type):

    def __instancecheck__(cls, obj: Any) -> bool:
        return obj is None or isinstance(obj, (str, int, float, bool))

    def __subclasscheck__(cls, t: type) -> bool:
        if t == JSONValueLike:
            return True
        return t is None or t == NoneType or issubclass(t, (str, int, float, bool))

class JSONValueLike(metaclass=_JSONValueLike):
    ...


class _JSONArrayLike(type):
    ALLOWED = (list, Sequence, collections.abc.Sequence)

    def __instancecheck__(cls, obj: Any) -> bool:
        return obj is not None and isinstance(obj, cls.ALLOWED) and \
            not isinstance(obj, str) and \
            all(isinstance(elem, JSONLike) for elem in obj)

    def __subclasscheck__(cls, t: type) -> bool:
        if t == JSONArrayLike:
            return True
        origin = get_origin(t)
        args = get_args(t)
        return ((inspect.isclass(origin) and issubclass(origin, cls.ALLOWED))
                or origin in cls.ALLOWED) and \
            not issubclass(origin, str) and \
            len(args) == 1 and issubclass(args[0], JSONLike)

class JSONArrayLike(metaclass=_JSONArrayLike):
    ...


class _JSONContainerLike(type):
    def __instancecheck__(cls, obj: Any) -> bool:
        return isinstance(obj, (JSONObjectLike, JSONArrayLike))

    def __subclasscheck__(self, t: type) -> bool:
        if t in (JSONContainerLike, JSONObjectLike, JSONArrayLike):
            return True
        return issubclass(t, (JSONObjectLike, JSONArrayLike))

class JSONContainerLike(metaclass=_JSONContainerLike):
    ...


class _JSONLike(type):
    def __instancecheck__(self, obj: Any) -> bool:
        return isinstance(obj, (JSONValueLike, JSONObjectLike, JSONArrayLike))

    def __subclasscheck__(cls, t: type) -> bool:
        if t in (JSONLike, JSONObjectLike, JSONArrayLike, JSONValueLike, JSONContainerLike):
            return True
        return issubclass(t, (JSONValueLike, JSONObjectLike, JSONArrayLike))

class JSONLike(metaclass=_JSONLike):
    ...


#


def is_json_type(value: Any, bound: JSONValue | JSONObject | JSONArray | JSONContainer | JSONType = JSONType) -> bool:
    if value is None:
        return bound in (JSONValue, JSONType)
    if isinstance(value, (bool, int, float, str)):
        return bound in (JSONValue, JSONType)
    if isinstance(value, list):
        return bound in (JSONArray, JSONContainer, JSONType) and (is_json_type(item) for item in value)
    if isinstance(value, dict):
        return bound in (JSONObject, JSONContainer, JSONType) and (isinstance(k, str) and is_json_type(v)
                                                                   for k, v in value.items())
    return False


class JSONable(ABC):

    @abstractmethod
    def to_json(self, **kwargs) -> JSONObject:
        ...


class JSONConvertible(JSONable, ABC):

    @classmethod
    @abstractmethod
    def from_json(cls, o: JSONObject, **kwargs) -> Self:
        ...



from __future__ import annotations

from SprelfJSON import JSONObjectLike, JSONValueLike
from SprelfJSON.JSONDefinitions import JSONConvertible, JSONable, JSONType, JSONLike, JSONArrayLike, JSONContainerLike
from SprelfJSON.Helpers import ClassHelpers, TimeHelpers
from SprelfJSON.JSONModel.JSONModelError import JSONModelError
from SprelfJSON.Objects import Ephemeral

from typing import Any, TypeVar, Callable, Union, Optional
from types import UnionType, NoneType, GeneratorType
from collections.abc import Sequence, MutableSequence, Mapping, MutableMapping, MutableSet, Collection, \
    Iterator, Iterable, Generator, Set
import typing_inspect
import inspect
import base64
import json
import re
from abc import ABC, abstractmethod

from datetime import datetime, date, time, timedelta
from enum import Enum, StrEnum, IntEnum, IntFlag

T2 = TypeVar('T2')
SupportedTypes = (dict, list, set, tuple, bool, str, int, float, bytes, type, None,
                  datetime, date, time, timedelta, re.Pattern,
                  Enum, StrEnum, IntEnum, IntFlag,
                  Sequence, MutableSequence, Mapping, MutableMapping, MutableSet, Collection,
                  Iterable, Iterator, Generator,
                  JSONable, UnionType, NoneType,
                  JSONValueLike, JSONContainerLike, JSONLike, JSONObjectLike, JSONArrayLike,
                  Ephemeral)
Supported = Union[SupportedTypes]
SupportedTypeMap = {t.__name__: t for t in SupportedTypes if t is not None}
_SupportedTypes_O1 = set(SupportedTypes)


class ModelElemError(JSONModelError):

    def __init__(self, model_elem: _BaseModelElem, *args):
        super().__init__(*args)
        self.model_elem = model_elem


#


class _BaseModelElem:
    """
    Base definition for a ModelElem.  Contains all elements related to storing a particular
    type, and parsing/dumping values of that type.  Not intended to be used directly.
    """
    __base64_altchars__: tuple[bytes, ...] = (b"-_", b"+/")
    _AliasedModelTypes: list[type[ModelType]] = None
    _ConcreteModelTypes: list[type[ModelType]] = None

    def __init__(self, typ: type[Supported], _ephemeral: bool = False, _type: bool = False):
        t, gen = type(self)._validate_definition(typ, _ephemeral, _type)
        self.origin: type[Supported] = t
        self.generics: tuple[_BaseModelElem, ...] = gen
        self._model_type: type[ModelType] | None = None

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return f"{type(self).__name__}({self.origin.__name__})"

    def is_generic(self) -> bool:
        return len(self.generics) > 0

    def is_union(self) -> bool:
        return self.origin == Union or typing_inspect.is_union_type(self.origin)

    @property
    def T(self) -> type[T]:
        return self.origin

    @property
    def annotated_type(self) -> type[Supported]:
        return ClassHelpers.as_generic(self.origin, *(g.annotated_type for g in self.generics))

    @property
    def annotation(self) -> str:
        if self.is_union():
            return "|".join(g.annotation for g in self.generics)
        suffix = "" if len(self.generics) == 0 else f"[{','.join(g.annotation for g in self.generics)}]"
        return f"{self.origin.__name__}{suffix}"

    def is_valid(self, value: Any, *, key: str | None = None, **kwargs) -> bool:
        try:
            _ = self.validate(value, key=key, **kwargs)
            return True
        except ModelElemError:
            return False

    def validate(self, value: Any, *, key: str | None = None, **kwargs) -> Supported:
        """
        Validates that the given value conforms to the type defined by this model element, transforming
        it if needed, and returning that potentially-transformed value.
        """
        parsed = self.parse_value(value, key=key, **kwargs)
        if self._is_valid(parsed, key=key, **kwargs):
            return parsed

        if isinstance(value, list) or isinstance(value, set):
            t_str = f"{type(value).__name__}[{'|'.join({type(v).__name__ for v in value})}]"
        elif isinstance(value, Mapping):
            t_str = f"dict[{'|'.join({type(k).__name__ for k in value.keys()})}, " \
                    f"{'|'.join({type(v).__name__ for v in value.values()})}]"
        else:
            t_str = type(value).__name__

        raise ModelElemError(self, f"Schema mismatch: Expected type '{self.annotation}', "
                                   f"but got '{t_str}' instead")

    def _is_valid(self, val: Supported, **kwargs) -> bool:
        return self.validate_type(val, **kwargs)

    #

    @classmethod
    def _resolve_model_types(cls):
        if cls._AliasedModelTypes is None or cls._ConcreteModelTypes is None:
            cls._AliasedModelTypes = [mt for mt in ClassHelpers.all_subclasses(ModelType)
                                      if not inspect.isabstract(mt)]
            cls._ConcreteModelTypes = [cls._AliasedModelTypes.pop(i)
                                       for i, mt in reversed(list(enumerate(cls._AliasedModelTypes)))
                                       if issubclass(mt, ModelType_Concrete)]

    def get_matching_model_type(self, **kwargs) -> type[ModelType]:
        if self._model_type is None:
            self._resolve_model_types()
            for mt in self._AliasedModelTypes:
                if mt.test_origin(self, **kwargs):
                    self._model_type = mt
                    return mt
            if inspect.isclass(self.origin):
                for mt in self._ConcreteModelTypes:
                    if mt.test_origin(self, **kwargs):
                        self._model_type = mt
                        return mt

            raise ModelElemError(self, f"Unable to identify matching ModelType for ModelElem with origin "
                                       f"of type '{self.annotated_type!r}'")

        return self._model_type

    #

    def validate_type(self, val: Supported, **kwargs) -> bool:
        try:
            mt = self.get_matching_model_type(**kwargs)
            return mt.is_valid(val, self, **kwargs)
        except ModelElemError:
            return False

    def parse_value(self, val: Any, **kwargs) -> Supported:
        """
        Parses the given value to conform with the type defined by this model element, if possible.
        If unsuccessful, a ModelElemError is raised.
        """
        return self._parse_value(val, **kwargs)

    def _parse_value(self, val: Any, **kwargs) -> Supported:
        mt = self.get_matching_model_type(**kwargs)
        return mt.parse(val, self, **kwargs)

    #

    def dump_value(self, val: Supported, *, key: str | None = None, **kwargs) -> JSONType:
        """
        Dumps the given value into a JSON-friendly type representing the type
        defined by this model element, if the given value can be validated (see validate()).
        If unable to validate, or unable to dump the validated value, a
        ModelElemError is raised.

        :keyword key: The key of the JSONModel that this ModelElem represents.  Just used for logging,
        not required.
        """
        return self._dump_value(val, key=key)

    def _dump_value(self, val: Supported, **kwargs) -> JSONType:
        mt = self.get_matching_model_type(**kwargs)
        return mt.dump(val, self, **kwargs)

    #

    @classmethod
    def _validate_definition(cls, val_type: type, _ephemeral: bool, _type: bool) -> \
            tuple[type, tuple[_BaseModelElem, ...]]:
        t, gen = ClassHelpers.analyze_type(val_type)
        if t is None or (inspect.isclass(t) and not _ephemeral and not _type and
                         all(not inspect.isclass(supported) or not issubclass(t, supported)
                             for supported in SupportedTypes)):
            raise JSONModelError(f"Cannot define ModelElem with unsupported type '{t.__name__}'.")
        if len(gen) == 0:
            return t, ()

        if t == type:
            return t, tuple(_BaseModelElem(arg, _type=True) for arg in gen)

        if inspect.isclass(t) and issubclass(t, dict) and len(gen) != 2:
            raise JSONModelError(f"Invalid dict definition for ModelElem: [{','.join(g.__name__ for g in gen)}]")

        if inspect.isclass(t) and issubclass(t, Ephemeral):
            return t, tuple(_BaseModelElem(arg, _ephemeral=True) for arg in gen)

        return t, tuple(_BaseModelElem(arg) for arg in gen)


#


#


class ModelElem(_BaseModelElem):
    """
    Full implementation of ModelElem, implementing extra features over
    _BaseModelElem, including default values and alternate parsing.
    """

    def __init__(self, typ: type[Supported],
                 *,
                 default: Union[Supported, tuple[()], None] = (),
                 default_factory: Callable[[], Supported] | None = None,
                 alternates: Iterable[AlternateModelElem] = (),
                 use_alternates_only: bool = False,
                 ignored: bool = False):
        super().__init__(typ)
        self._ignored = ignored
        self._is_ephemeral = inspect.isclass(self.origin) and issubclass(self.origin, Ephemeral)
        self._alternates = list(alternates)
        self._use_alternates_only = use_alternates_only
        if default_factory is not None:
            self._default_factory = default_factory
            self._default = ()
        elif self._is_ephemeral:
            self._default_factory = None
            self._default = (default.value,) if isinstance(default, Ephemeral) \
                else (default,) if default != () else (None,)
        else:
            self._default_factory = None
            self._default: tuple[Optional[Supported]] = \
                (default,) if not isinstance(default, tuple) or len(default) > 0 else default

    # OVERRIDE
    def __str__(self) -> str:
        values = [self.origin.__name__]
        if not isinstance(self.default, tuple) or len(self.default) > 0:
            values.append(f"default={self.default!r}")
        return f"{type(self).__name__}({','.join(values)})"

    @property
    def default(self) -> Optional[Supported]:
        if not self.has_default():
            raise ModelElemError(self, "ModelElem does not have a default value; cannot access")
        if self._default_factory is not None:
            return self._default_factory()
        return self._default[0]

    def has_default(self) -> bool:
        return len(self._default) > 0 or self._default_factory is not None

    @property
    def ignored(self) -> bool:
        return self._ignored

    @property
    def ephemeral(self) -> bool:
        return self._is_ephemeral

    #

    # OVERRIDE
    def parse_value(self, val: Any, *, key: str | None = None, **kwargs) -> Supported:
        if self.ignored:
            return None

        if not self._use_alternates_only:
            try:
                return self._parse_value(val, key=key, **kwargs)
            except:
                if len(self._alternates) == 0:
                    raise

        for a in self._alternates:
            try:
                return a.transformer(a.parse_value(val, key=key, **kwargs))
            except:
                continue
        raise ModelElemError(self, f"Unable to parse value of type '{type(val).__name__}' as "
                                   f"an object of type '{self.annotated_type!r}'" +
                             (f" ({len(self._alternates)} alternate(s) also failed)"
                              if len(self._alternates) > 0 else "") +
                             ".")

    # OVERRIDE
    def dump_value(self, val: Supported, *, key: str | None = None, **kwargs) -> JSONType:
        if self.ignored:
            return None
        if self.ephemeral:
            raise ModelElemError(self, "Cannot dump ephemeral object.")

        if not self._use_alternates_only:
            try:
                return self._dump_value(val, key=key, **kwargs)
            except:
                if len(self._alternates) == 0:
                    raise

        for a in self._alternates:
            if a.jsonifier is None:
                continue
            try:
                return a.jsonifier(val)
            except:
                continue

        raise ModelElemError(self, f"Unable to dump value of type '{type(val).__name__}' as "
                                   f"an object of type '{self.annotated_type!r}'" +
                             (f" ({len(self._alternates)} alternates also failed)"
                              if len(self._alternates) > 0 else "") +
                             ".")

    # OVERRIDE
    def validate(self, value: Any, **kwargs) -> Supported:
        if self.ignored:
            return None

        return super().validate(value, **kwargs)


#


#


class AlternateModelElem(_BaseModelElem):

    def __init__(self, typ: type[T2],
                 transformer: Callable[[T2], Supported],
                 jsonifier: Callable[[Any], SupportedTypes] | None = None):
        super().__init__(typ)
        self.transformer = transformer
        self.jsonifier = jsonifier


#


#


#


class ModelType(ABC):

    @classmethod
    @abstractmethod
    def test_origin(cls, elem: _BaseModelElem, **kwargs) -> bool:
        pass

    @classmethod
    def is_valid(cls, val: Supported, elem: _BaseModelElem, **kwargs) -> bool:
        return isinstance(val, elem.origin)

    @classmethod
    @abstractmethod
    def parse(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        pass

    @classmethod
    @abstractmethod
    def dump(cls, val: Any, elem: _BaseModelElem, **kwargs) -> JSONType:
        pass

    @classmethod
    def _parse_error(cls, val: Any, elem: _BaseModelElem, message: str, **kwargs):
        k_str = f" on key '{k}'" if (k := kwargs.pop("key", None)) else ""
        return ModelElemError(elem, f"Unable to parse value of type '{type(val).__name__}'{k_str} as "
                                    f"type '{elem.annotation}'" + (message or "."))

    @classmethod
    def _dump_error(cls, val: Any, elem: _BaseModelElem, message: str, **kwargs) -> ModelElemError:
        k_str = f" on key '{k}'" if (k := kwargs.pop("key", None)) else ""
        return ModelElemError(elem, f"Given value of type '{type(val).__name__}'{k_str}) " + message)

    @classmethod
    def _parse_for_dump(cls, val: Any, elem: _BaseModelElem, **kwargs) -> Any:
        try:
            return elem.validate(val)
        except ModelElemError as e:
            raise cls._dump_error(val, elem, f"as the expected type '{elem.annotated_type!r}'; "
                                             f"could not be validated: {str(e)}", **kwargs)


#


# noinspection PyPep8Naming
class ModelType_None(ModelType):

    @classmethod
    def test_origin(cls, elem: _BaseModelElem, **kwargs) -> bool:
        return elem.origin is None or elem.origin is NoneType

    @classmethod
    def is_valid(cls, val: Supported, elem: _BaseModelElem, **kwargs) -> bool:
        return val is None

    @classmethod
    def parse(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        return None

    @classmethod
    def dump(cls, val: Any, elem: _BaseModelElem, **kwargs) -> JSONType:
        return None


# noinspection PyPep8Naming
class ModelType_JSONLike(ModelType):

    @classmethod
    def test_origin(cls, elem: _BaseModelElem, **kwargs) -> bool:
        return elem.origin in (JSONLike, JSONObjectLike, JSONArrayLike, JSONValueLike, JSONContainerLike)

    @classmethod
    def parse(cls, val: Any, elem: _BaseModelElem, **kwargs):
        if not isinstance(val, elem.origin):
            raise cls._parse_error(val, elem, f"; does not fit the specified JSON-compatible structure.")
        return val

    @classmethod
    def dump(cls, val: Any, elem: _BaseModelElem, **kwargs) -> JSONType:
        if not isinstance(val, elem.origin):
            raise cls._dump_error(val, elem, f"could not be dumped as specified JSON-compatible structure.")
        return val


# noinspection PyPep8Naming
class ModelType_Union(ModelType):

    @classmethod
    def test_origin(cls, elem: _BaseModelElem, **kwargs) -> bool:
        return typing_inspect.is_union_type(elem.origin) or elem.origin in (Union, UnionType)

    @classmethod
    def is_valid(cls, val: Supported, elem: _BaseModelElem, **kwargs) -> bool:
        return any(g.validate_type(val) for g in elem.generics)

    @classmethod
    def parse(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        for g in elem.generics:
            try:
                return g.parse_value(val)
            except ModelElemError:
                pass
        raise ModelElemError(elem, f"Given value of type '{type(val).__name__}' does not meet any of the allowed "
                                   f"union types: {', '.join(repr(g.annotated_type) for g in elem.generics)}")

    @classmethod
    def dump(cls, val: Any, elem: _BaseModelElem, **kwargs) -> JSONType:
        for g in elem.generics:
            try:
                return g.dump_value(val, **kwargs)
            except ModelElemError:
                pass
        raise cls._dump_error(val, elem, "cannot be dumped as any of the allowed "
                                         f"types: {', '.join(repr(g.annotated_type) for g in elem.generics)}",
                              **kwargs)


# noinspection PyPep8Naming
class ModelType_Optional(ModelType):

    @classmethod
    def test_origin(cls, elem: _BaseModelElem, **kwargs) -> bool:
        return typing_inspect.is_optional_type(elem.origin)

    @classmethod
    def parse(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        if val is None:
            return val
        if len(elem.generics) > 0:
            return elem.generics[0].parse_value(val)
        return val

    @classmethod
    def is_valid(cls, val: Supported, elem: _BaseModelElem, **kwargs) -> bool:
        if val is None:
            return True
        return len(elem.generics) == 0 or elem.generics[0].is_valid(val)

    @classmethod
    def dump(cls, val: Any, elem: _BaseModelElem, **kwargs) -> JSONType:
        if val is None:
            return None
        if len(elem.generics) > 0:
            return elem.generics[0].dump_value(val, **kwargs)


# noinspection PyPep8Naming
class ModelType_Generator(ModelType):

    @classmethod
    def test_origin(cls, elem: _BaseModelElem, **kwargs) -> bool:
        return elem.origin in (Generator, Iterable, Iterator, GeneratorType)

    @classmethod
    def parse(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        if not isinstance(val, Iterable):
            raise cls._parse_error(val, elem, f"; value is not iterable.")
        if len(elem.generics) > 0:
            return (elem.generics[0].dump_value(v) for v in val)  # Stays lazy
        return val

    @classmethod
    def dump(cls, val: Any, elem: _BaseModelElem, **kwargs) -> JSONType:
        parsed = cls._parse_for_dump(val, elem, **kwargs)
        if not isinstance(parsed, Iterable):
            raise cls._dump_error(val, elem, f"is not iterable; cannot dump as a JSON array.", **kwargs)
        if len(elem.generics) > 0:
            return [elem.generics[0].dump_value(v) for v in parsed]
        return list(parsed)


# noinspection PyPep8Naming
class ModelType_Sequence(ModelType_Generator):

    @classmethod
    def test_origin(cls, elem: _BaseModelElem, **kwargs) -> bool:
        return elem.origin in (Sequence, Collection, MutableSequence, MutableSet, Set)

    @classmethod
    def is_valid(cls, val: Supported, elem: _BaseModelElem, **kwargs) -> bool:
        return (isinstance(val, elem.origin)) \
            and (len(elem.generics) == 0 or all(elem.generics[0].is_valid(x) for x in val))

    @classmethod
    def parse(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        if not isinstance(val, Iterable):
            raise cls._parse_error(val, elem, f"; value is not an iterable.")
        if len(elem.generics) > 0:
            val = (elem.generics[0].parse_value(v) for v in val)
        if elem.origin in (Sequence, Collection, MutableSequence):
            return list(val)
        elif elem.origin in (MutableSet, Set):
            return set(val)
        raise ModelElemError(elem, f"Unable to find suitable type to parse value as given "
                                   f"the expected type '{elem.annotated_type!r}'")


# noinspection PyPep8Naming
class ModelType_Object(ModelType, ABC):
    t: type

    @classmethod
    def test_origin(cls, elem: _BaseModelElem, **kwargs) -> bool:
        try:
            return issubclass(elem.origin, cls.t)
        except TypeError:
            return False


# noinspection PyPep8Naming
class ModelType_List(ModelType_Object):
    t = list

    @classmethod
    def is_valid(cls, val: Supported, elem: _BaseModelElem, **kwargs) -> bool:
        return (isinstance(val, cls.t) and
                (len(elem.generics) == 0 or all(elem.generics[0].is_valid(x, **kwargs)
                                                for x in val)))

    @classmethod
    def parse(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        if not isinstance(val, Iterable):
            raise cls._parse_error(val, elem, f"; object is not iterable.", **kwargs)
        if len(elem.generics) > 0:
            return cls.t(elem.generics[0].parse_value(v, **kwargs) for v in val)
        return cls.t(val)

    @classmethod
    def dump(cls, val: Any, elem: _BaseModelElem, **kwargs) -> JSONType:
        parsed = cls._parse_for_dump(val, elem, **kwargs)
        if not isinstance(parsed, Iterable):
            raise cls._dump_error(val, elem, f"as an iterable; "
                                             f"cannot dump as a JSON array.", **kwargs)
        if len(elem.generics) > 0:
            return [elem.generics[0].dump_value(v) for v in parsed]
        return list(parsed)


# noinspection PyPep8Naming
class ModelType_Set(ModelType_List):
    t = set


# noinspection PyPep8Naming
class ModelType_FrozenSet(ModelType_List):
    t = frozenset


# noinspection PyPep8Naming
class ModelType_Tuple(ModelType_List):
    t = tuple

    @classmethod
    def is_valid(cls, val: Supported, elem: _BaseModelElem, **kwargs) -> bool:
        if not isinstance(val, tuple):
            return False
        if len(elem.generics) == 2 and elem.generics[1].origin is Ellipsis:
            return all(elem.generics[0].is_valid(x) for x in val)
        if len(elem.generics) == 0:
            return True
        if len(elem.generics) != len(val):
            return False
        return all(g.is_valid(x) for g, x in zip(elem.generics, val))

    @classmethod
    def parse(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        if not isinstance(val, Collection):
            raise cls._parse_error(val, elem, f"; value is not a collection.", **kwargs)
        if len(elem.generics) == 2 and elem.generics[1].origin is Ellipsis:
            return cls.t(elem.generics[0].parse_value(v) for v in val)
        elif len(elem.generics) == 0:
            return cls.t(val)
        elif len(elem.generics) == len(val):
            return cls.t(g.parse_value(v) for g, v in zip(elem.generics, val))
        raise cls._parse_error(val, elem, f"; has the wrong number of "
                                          f"elements to be parsed as a '{elem.annotated_type!r}'; has ({len(val)}).",
                               **kwargs)

    @classmethod
    def dump(cls, val: Any, elem: _BaseModelElem, **kwargs) -> JSONType:
        if len(elem.generics) == 2 and elem.generics[1].origin is Ellipsis:
            return super().dump(val, elem, **kwargs)

        parsed = cls._parse_for_dump(val, elem, **kwargs)
        if not isinstance(parsed, Iterable):
            raise cls._dump_error(val, elem, f" is not iterable; "
                                             f"cannot dump as a JSON array.", **kwargs)
        if len(elem.generics) == 0:
            return list(parsed)
        if len(val) == len(elem.generics):
            return [g.dump_value(x) for x, g in zip(val, elem.generics)]
        raise cls._parse_error(val, elem, f"; has the wrong number of "
                                          f"elements to be dumped as a '{elem.annotated_type!r}'; has ({len(val)}).",
                               **kwargs)


# noinspection PyPep8Naming
class ModelType_Dict(ModelType):

    @classmethod
    def test_origin(cls, elem: _BaseModelElem, **kwargs) -> bool:
        return elem.origin in (Mapping, MutableMapping) \
            or (inspect.isclass(elem.origin) and issubclass(elem.origin, dict))

    @classmethod
    def is_valid(cls, val: Supported, elem: _BaseModelElem, **kwargs) -> bool:
        return (isinstance(val, Mapping) and
                (len(elem.generics) == 0 or all(elem.generics[0].is_valid(k) and elem.generics[1].is_valid(v)
                                                for k, v in val.items())))

    @classmethod
    def parse(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        if ClassHelpers.check_generic_instance(val, Iterable, tuple[Any, Any]):
            d = {k: v for k, v in val}
        else:
            d = val
        if isinstance(d, Mapping):
            if len(elem.generics) == 0:
                return d
            if all(isinstance(k, str) for k in d.keys()):
                if inspect.isclass(elem.generics[0].origin) and issubclass(elem.generics[0].origin, str):
                    return {str(k): elem.generics[1].parse_value(v) for k, v in d.items()}
                return {elem.generics[0].parse_value(json.loads(k)): elem.generics[1].parse_value(v)
                        for k, v in d.items()}
            else:
                return {elem.generics[0].parse_value(k): elem.generics[1].parse_value(v)
                        for k, v in d.items()}
        raise cls._parse_error(val, elem, f"", **kwargs)

    @classmethod
    def dump(cls, val: Any, elem: _BaseModelElem, **kwargs) -> JSONType:
        parsed = cls._parse_for_dump(val, elem, **kwargs)
        if len(elem.generics) == 0:
            return dict(parsed)
        if inspect.isclass(elem.generics[0].origin):
            if issubclass(elem.generics[0].origin, str):
                return {str(k): elem.generics[1].dump_value(v) for k, v in parsed.items()}

        return {(json.dumps(dk) if not isinstance(dk, str) else dk): dv
                for dk, dv in
                ((elem.generics[0].dump_value(k), elem.generics[1].dump_value(v)) for k, v in parsed.items())}


# noinspection PyPep8Naming
class ModelType_Type(ModelType_Object):
    t = type
    __output_full_class_name__: bool = True

    @classmethod
    def is_valid(cls, val: Supported, elem: _BaseModelElem, **kwargs) -> bool:
        return (inspect.isclass(val) and
                (len(elem.generics) == 0 or ClassHelpers.check_subclass(val, elem.generics[0].annotated_type)))

    @classmethod
    def parse(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        if isinstance(val, str):
            if val.lower() == "datetime":
                return datetime
            elif val.lower() == "date":
                return date
            elif val.lower() == "time":
                return time
            elif val.lower() in ("null", "none"):
                return None
            t = ClassHelpers.locate_class(val)
            if t is None:
                raise ModelElemError(elem, f"Unable to parse string value '{val}' as a type; "
                                           f"type is not found.")
            val = t
        if not inspect.isclass(val):
            raise cls._parse_error(val, elem, f"; could not be "
                                              f"interpreted as a type.  Is it an instance?", **kwargs)
        return val

    @classmethod
    def dump(cls, val: Any, elem: _BaseModelElem, **kwargs) -> JSONType:
        parsed = cls._parse_for_dump(val, elem, **kwargs)
        if parsed is None:
            return None
        if any(parsed.__module__ == m for m in ("builtins", "datetime", "typing", "enum")) \
                or not cls.__output_full_class_name__:
            return parsed.__name__
        return ClassHelpers.full_name(parsed)


# noinspection PyPep8Naming
class ModelType_Concrete(ModelType_Object, ABC):

    @classmethod
    def parse(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        if isinstance(val, elem.origin):
            return val
        else:
            return cls._convert(val, elem, **kwargs)

    @classmethod
    @abstractmethod
    def _convert(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        ...


# noinspection PyPep8Naming
class ModelType_Basic(ModelType_Concrete):

    @classmethod
    def test_origin(cls, elem: _BaseModelElem, **kwargs) -> bool:
        return any(issubclass(elem.origin, x) for x in (str, int, float, bool))

    @classmethod
    def _convert(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        if issubclass(elem.origin, float) and isinstance(val, int):
            return float(val)
        raise cls._parse_error(val, elem, "", **kwargs)

    @classmethod
    def dump(cls, val: Any, elem: _BaseModelElem, **kwargs) -> JSONType:
        return cls._parse_for_dump(val, elem, **kwargs)


# noinspection PyPep8Naming
class ModelType_Pattern(ModelType_Concrete):

    @classmethod
    def test_origin(cls, elem: _BaseModelElem, **kwargs) -> bool:
        return issubclass(elem.origin, re.Pattern)

    @classmethod
    def _convert(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        if isinstance(val, str):
            try:
                return re.compile(val)
            except:
                raise ModelElemError(elem, f"Unable to compile string as a regular expression.")
        raise cls._parse_error(val, elem, f"", **kwargs)

    @classmethod
    def dump(cls, val: Any, elem: _BaseModelElem, **kwargs) -> JSONType:
        parsed = cls._parse_for_dump(val, elem, **kwargs)
        return parsed.pattern


# noinspection PyPep8Naming
class ModelType_DateTime(ModelType_Concrete):
    t = datetime

    @classmethod
    def test_origin(cls, elem: _BaseModelElem, **kwargs) -> bool:
        return elem.origin == datetime

    @classmethod
    def _convert(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        try:
            return TimeHelpers.parse_datetime(val)
        except ValueError:
            raise cls._parse_error(val, elem, "", **kwargs)

    @classmethod
    def dump(cls, val: Any, elem: _BaseModelElem, **kwargs) -> JSONType:
        return TimeHelpers.stringify_datetime(cls._parse_for_dump(val, elem, **kwargs))


# noinspection PyPep8Naming
class ModelType_Date(ModelType_Concrete):
    t = date

    @classmethod
    def test_origin(cls, elem: _BaseModelElem, **kwargs) -> bool:
        return elem.origin == date

    @classmethod
    def _convert(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        try:
            return TimeHelpers.parse_date(val)
        except ValueError:
            raise cls._parse_error(val, elem, "", **kwargs)

    @classmethod
    def dump(cls, val: Any, elem: _BaseModelElem, **kwargs) -> JSONType:
        return TimeHelpers.stringify_date(cls._parse_for_dump(val, elem, **kwargs))


# noinspection PyPep8Naming
class ModelType_Time(ModelType_Concrete):
    t = time

    @classmethod
    def test_origin(cls, elem: _BaseModelElem, **kwargs) -> bool:
        return elem.origin == time

    @classmethod
    def _convert(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        try:
            return TimeHelpers.parse_time(val)
        except ValueError:
            raise cls._parse_error(val, elem, "", **kwargs)

    @classmethod
    def dump(cls, val: Any, elem: _BaseModelElem, **kwargs) -> JSONType:
        return TimeHelpers.stringify_time(cls._parse_for_dump(val, elem, **kwargs))


# noinspection PyPep8Naming
class ModelType_TimeDelta(ModelType_Concrete):
    t = timedelta

    @classmethod
    def _convert(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        if isinstance(val, str):
            if m := TimeHelpers.TIMEDELTA_REGEX.match(val):
                groups = m.groupdict()
                try:
                    return timedelta(days=int(groups.get("d", 0)) or 0,
                                     hours=int(groups.get("h", 0)) or 0,
                                     minutes=int(groups.get("m", 0)) or 0,
                                     seconds=int(groups.get("s", 0)) or 0,
                                     milliseconds=int(groups.get("ms", 0)) or 0)
                except ValueError:
                    pass
            raise ModelElemError(elem, f"Unable to parse timedelta string; format is invalid.")
        elif isinstance(val, int):
            return timedelta(seconds=val / 1000)
        elif isinstance(val, float):
            return timedelta(seconds=val)
        raise cls._parse_error(val, elem, "", **kwargs)

    @classmethod
    def dump(cls, val: Any, elem: _BaseModelElem, **kwargs) -> JSONType:
        return int(cls._parse_for_dump(val, elem, **kwargs).total_seconds() * 1000)


# noinspection PyPep8Naming
class ModelType_IntFlag(ModelType_Concrete):
    t = IntFlag

    @classmethod
    def _convert(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        if isinstance(val, int):
            try:
                return elem.origin(val)
            except ValueError as e:
                raise ModelElemError(elem, f"Unable to parse IntFlag: {e}")
        raise cls._parse_error(val, elem, "", **kwargs)

    @classmethod
    def dump(cls, val: Any, elem: _BaseModelElem, **kwargs) -> JSONType:
        return cls._parse_for_dump(val, elem, **kwargs).value


# noinspection PyPep8Naming
class ModelType_IntEnum(ModelType_Concrete):
    t = IntEnum

    @classmethod
    def _convert(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        if isinstance(val, int):
            try:
                return elem.origin(val)
            except ValueError:
                raise ModelElemError(elem, f"Unable to parse IntEnum from integer; "
                                           f"Unrecognized IntEnum value ({val}) for '{type(elem.origin).__name__}'.")
        elif isinstance(val, str):
            try:
                return elem.origin[val]
            except KeyError:
                raise ModelElemError(elem, f"Unable to parse IntEnum from string; "
                                           f"Unrecognized IntEnum value ({val}) for '{type(elem.origin).__name__}'.")

        raise cls._parse_error(val, elem, "", **kwargs)

    @classmethod
    def dump(cls, val: Any, elem: _BaseModelElem, **kwargs) -> JSONType:
        return cls._parse_for_dump(val, elem, **kwargs).value


# noinspection PyPep8Naming
class ModelType_StrEnum(ModelType_Concrete):
    t = StrEnum

    @classmethod
    def _convert(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        if isinstance(val, str):
            try:
                return elem.origin(val)
            except ValueError:
                raise ModelElemError(elem, f"Unable to parse StrEnum; "
                                           f"Unrecognized StrEnum value ({val}) for '{type(elem.origin).__name__}")
        raise cls._parse_error(val, elem, "", **kwargs)

    @classmethod
    def dump(cls, val: Any, elem: _BaseModelElem, **kwargs) -> JSONType:
        return cls._parse_for_dump(val, elem, **kwargs).value


# noinspection PyPep8Naming
class ModelType_Enum(ModelType_Concrete):
    t = Enum

    @classmethod
    def test_origin(cls, elem: _BaseModelElem, **kwargs) -> bool:
        return issubclass(elem.origin, Enum) and not any(issubclass(elem.origin, x)
                                                         for x in (IntEnum, StrEnum, IntFlag))

    @classmethod
    def _convert(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        if isinstance(val, str):
            try:
                return elem.origin[val]
            except KeyError:
                pass
        try:
            return elem.origin(val)
        except ValueError:
            raise cls._parse_error(val, elem, "", **kwargs)

    @classmethod
    def dump(cls, val: Any, elem: _BaseModelElem, **kwargs) -> JSONType:
        return cls._parse_for_dump(val, elem, **kwargs).name


# noinspection PyPep8Naming
class ModelType_JSON(ModelType_Concrete):
    t = JSONConvertible

    @classmethod
    def _convert(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        if isinstance(val, str):
            try:
                return elem.origin.from_json(json.loads(val))
            except json.JSONDecodeError as e:
                raise cls._parse_error(val, elem, f"; failed to parse JSON from string: {str(e)}", **kwargs)
        if ClassHelpers.check_generic_instance(val, dict, str, Any):
            return elem.origin.from_json(val)
        raise cls._parse_error(val, elem, f"",
                               **kwargs)

    @classmethod
    def dump(cls, val: Any, elem: _BaseModelElem, **kwargs) -> JSONType:
        return cls._parse_for_dump(val, elem, **kwargs).to_json()


# noinspection PyPep8Naming
class ModelType_Bytes(ModelType_Concrete):
    t = bytes

    @classmethod
    def _convert(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        if isinstance(val, str):
            for alt in type(elem).__base64_altchars__:
                try:
                    return base64.b64decode(val, alt)
                except ValueError:
                    pass
            raise ModelElemError(elem, f"Unable to parse string to bytes; is not valid base-64")
        if ClassHelpers.check_generic_instance(val, Collection, int):
            if all(0 <= x <= 255 for x in val):
                return bytes(val)
            raise ModelElemError(elem, f"Given array of integers to parse as bytes has one or more values "
                                       f"outside the range of 0-255.")
        if ClassHelpers.check_generic_instance(val, Collection, str):
            try:
                parsed = [int(s, 16) for s in val]
            except ValueError:
                raise ModelElemError(elem, f"Given array of strings to parse as bytes has one or more invalid "
                                           f"hexadecimal strings.")
            if all(0 <= x <= 255 for x in parsed):
                return bytes(parsed)
            raise ModelElemError(elem, f"Given array of strings to parse as bytes has one or more hexadecimal "
                                       f"values outside the range of 0-255.")

    @classmethod
    def dump(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        parsed = cls._parse_for_dump(val, elem, **kwargs)
        alts = type(elem).__base64_altchars__
        alt = alts[0] if len(alts) > 0 else None
        return base64.b64encode(parsed, alt).decode("ascii")


# noinspection PyPep8Naming
class ModelType_Ephemeral(ModelType_Object):
    t = Ephemeral

    @classmethod
    def is_valid(cls, val: Supported, elem: _BaseModelElem, **kwargs) -> bool:
        if isinstance(val, Ephemeral):
            val = val.value
        if val is None:
            return True
        if not elem.is_generic():
            return True
        orig, gen = elem.generics[0].origin, elem.generics[0].generics
        if ClassHelpers.check_generic_instance(val, orig, *gen):
            return True
        return False

    @classmethod
    def parse(cls, val: Any, elem: _BaseModelElem, **kwargs) -> type[Supported]:
        if isinstance(val, Ephemeral):
            val = val.value
        if cls.is_valid(val, elem, **kwargs):
            return val
        raise cls._parse_error(val, elem, f"; must already be an object of the expected type.  Also, "
                                          f"beware of using custom generic classes, as they might not "
                                          f"be validated correctly here (TODO).  A workaround in this case "
                                          f"is to just not specific the generics in the definition.", **kwargs)

    @classmethod
    def dump(cls, val: Any, elem: _BaseModelElem, **kwargs) -> JSONType:
        return cls._parse_for_dump(val, elem, **kwargs)



from __future__ import annotations

import functools
from typing import Callable, TypeVar, Hashable, Iterable


THash = TypeVar('THash', bound=Hashable)


def yield_unique(func: Callable[..., Iterable[THash]]) -> Callable[..., Iterable[THash]]:

    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> Iterable[THash]:
        s = set()
        for x in func(*args, **kwargs):
            if x not in s:
                s.add(x)
                yield x

    return wrapper



from datetime import datetime, date, time, timezone
import re


DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"
DATETIME_FORMAT2 = "%Y-%m-%dT%H:%M:%S"
DATE_FORMAT = "%Y-%m-%d"
TIME_FORMAT = "%H:%M:%S.%f"
TIME_FORMAT2 = "%H:%M:%S"


TIMEDELTA_REGEX = re.compile('^(days=(?P<d>\d+),)?(hours=(?P<h>\d+),)?(minutes=(?P<m>\d+),)?seconds=(?P<s>\d+)(\.(?P<ms>\d+))?$')


def stringify_datetime(d: datetime) -> str:
    base = d.strftime(DATETIME_FORMAT2)
    millis = f"{round(d.microsecond / 1000):03d}"
    base = f"{base}.{millis}"
    if d.tzinfo is None or d.tzinfo == timezone.utc:
        return base + "Z"
    else:
        return base + d.strftime("%z")

def stringify_time(d: time) -> str:
    iso = d.strftime(TIME_FORMAT)
    decimal_pos = iso.rfind(".")
    seconds_fraction = iso[decimal_pos + 1:]
    millis = seconds_fraction.rjust(3, "0")[:3]
    return iso[:decimal_pos + 1] + millis

def stringify_date(d: date):
    return d.strftime(DATE_FORMAT)


def parse_datetime_string(s: str) -> datetime:
    try:
        dt = datetime.fromisoformat(s.strip("Z"))
        return dt if dt.tzinfo else dt.replace(tzinfo=timezone.utc)
    except ValueError:
        try:
            return datetime.strptime(s, DATETIME_FORMAT).replace(tzinfo=timezone.utc)
        except ValueError:
            return datetime.strptime(s, DATETIME_FORMAT2).replace(tzinfo=timezone.utc)


def parse_datetime(d: datetime | date | time | str | float | int) -> datetime:
    if isinstance(d, datetime):
        return d if d.tzinfo else d.replace(tzinfo=timezone.utc)
    if isinstance(d, str):
        return parse_datetime_string(d)
    if isinstance(d, int) or isinstance(d, float):
        return datetime.fromtimestamp(d, tz=timezone.utc)
    if isinstance(d, date):
        return datetime(year=d.year, month=d.month, day=d.day, tzinfo=timezone.utc)
    if isinstance(d, time):
        return datetime(year=datetime.min.year,
                        month=datetime.min.month,
                        day=datetime.min.day,
                        hour=d.hour, minute=d.minute,
                        second=d.second, microsecond=d.microsecond,
                        tzinfo=timezone.utc)
    raise ValueError(f"Unable to smart-parse value as datetime: {d}")


def parse_date_string(s: str) -> date:
    try:
        return datetime.fromisoformat(s.strip("Z")).date()
    except ValueError:
        try:
            return datetime.strptime(s, DATE_FORMAT).date()
        except ValueError:
            return parse_datetime_string(s).date()


def parse_date(d: datetime | date | str | float | int) -> date:
    if isinstance(d, date):
        return d
    if isinstance(d, str):
        return parse_date_string(d)
    if isinstance(d, int) or isinstance(d, float):
        return datetime.fromtimestamp(d, tz=timezone.utc).date()
    if isinstance(d, datetime):
        return date(year=d.year, month=d.month, day=d.day)
    raise ValueError(f"Unable to smart-parse value as date: {d}")


def parse_time_string(s: str) -> time:
    try:
        return datetime.fromisoformat(s.strip("Z")).time()
    except ValueError:
        try:
            return datetime.strptime(s, TIME_FORMAT).time()
        except ValueError:
            return datetime.strptime(s, TIME_FORMAT2).time()


def parse_time(t: datetime | time | str | float | int) -> time:
    if isinstance(t, time):
        return t
    if isinstance(t, str):
        return parse_time_string(t)
    if isinstance(t, int) or isinstance(t, float):
        return datetime.fromtimestamp(t, tz=timezone.utc).time()
    if isinstance(t, datetime):
        return t.time()
    raise ValueError(f"Unable to smart-parse value as time: {t}")


def day(dt: datetime) -> datetime:
    return datetime.combine(dt.date(), datetime.min.time())



from __future__ import annotations

from abc import ABCMeta, ABC, abstractmethod
from typing import Any, Hashable, Self, Mapping, dataclass_transform, TypeAlias
from types import resolve_bases, new_class, GenericAlias
import typing
import itertools
import inspect

import typing_inspect

from SprelfJSON import JSONArrayLike
from SprelfJSON.JSONModel.ModelElem import ModelElem, ModelElemError, SupportedTypeMap, AlternateModelElem
from SprelfJSON.JSONDefinitions import JSONObject, JSONType, JSONConvertible, JSONContainerLike, JSONValueLike, \
    JSONContainer, JSONValue, JSONLike, JSONObjectLike, JSONArray
from SprelfJSON.JSONModel.JSONModelError import JSONModelError
from SprelfJSON.Helpers import ClassHelpers


#


@dataclass_transform(kw_only_default=True, field_specifiers=(ModelElem,))
class JSONModelMeta(ABCMeta):
    _ANNO = '__annotations__'
    _FIELDS = "_fields"
    _JSON_MODEL = "__json_model__"
    _DEFAULTS = "__defaults__"
    __json_model__: dict[str, ModelElem]
    __name_field__: str = "__name"
    __name_field_required__: bool = False
    __include_name_in_json_output__: bool = False
    __allow_null_json_output__: bool = False
    __include_defaults_in_json_output__: bool = False
    __allow_extra_fields__: bool = False
    __exclusions__: list[str] = []
    __eval_context__ = {**globals(),
                        **SupportedTypeMap,
                        **typing.__dict__,
                        ModelElem.__name__: ModelElem,
                        AlternateModelElem.__name__: AlternateModelElem}
    __annotation_conversions__ = {JSONObject: JSONObjectLike,
                                  JSONArray: JSONArrayLike,
                                  JSONType: JSONLike,
                                  JSONContainer: JSONContainerLike,
                                  JSONValue: JSONValueLike}

    @classmethod
    def _eval(cls, s: Any, context: dict):
        if s is None or inspect.isclass(s) or typing_inspect.is_generic_type(s) or typing_inspect.is_union_type(s) or \
                typing_inspect.is_optional_type(s) or isinstance(s, ModelElem) or type(s) == GenericAlias:
            return s
        try:
            return eval(s, context)
        except:
            if c := ClassHelpers.locate_class(s):
                return c
            raise JSONModelError(f"Unable to evaluate '{s}' as a type; has it been instantiated yet?  Be sure "
                                 f"nested classes are defined before the classes they're nested in, and "
                                 f"avoid circular references.")

    #

    def __new__(mcls: type[JSONModelMeta],
                name: str,
                bases: tuple[type, ...],
                namespace: dict[str, Any],
                **kwargs) -> type:
        _super = super()

        def BUILD(_name, _bases, _namespace):
            # There are a lot of ways to consider building the object; some options are commented out.
            # return type.__new__(mcls, _name, _bases, _namespace)
            # return ABCMeta.__new__(mcls, _name, _bases, _namespace)
            # return new_class(_name, _bases,
            #                  exec_body=lambda ns: ns.update({**_namespace}))
            _bases = resolve_bases(_bases)
            return _super.__new__(mcls, _name, _bases, _namespace)

        #

        if name == "JSONModel":
            new_cls = BUILD(name, bases, namespace)
            return new_cls

        #
        # Retrieve all annotated values and defaults for the class being created and its parents
        #
        given_anno: dict[str, Any] = {}
        defaults: dict[str, Any] = {}

        for base in reversed(bases):
            if hasattr(base, mcls._ANNO):
                given_anno.update(base.__annotations__)
                defaults.update(getattr(base, mcls._DEFAULTS, {}))
            if base == ABC:
                namespace["__name_field_required__"] = True
                namespace["__include_name_in_json_output__"] = True

        given_anno.update(namespace.get(mcls._ANNO, {}))

        defaults.update({key: namespace[key] for key in given_anno.keys() if key in namespace})
        defaults.update({key: elem.default for key, elem in given_anno.items()
                         if isinstance(elem, ModelElem) and elem.has_default()})
        given_anno = {k: v for k, v in given_anno.items() if not k.startswith("_")}
        defaults = {k: v for k, v in defaults.items() if not k.startswith("_")}

        for field in given_anno.keys():
            if hasattr(mcls, field):
                defaults[field] = getattr(mcls, field)

        #
        # Fill in the namespace and build
        #
        namespace = {
            **namespace,
            mcls._DEFAULTS: defaults
        }

        new_cls = BUILD(name, bases, namespace)

        #
        # Parse the annotated strings to get actual types to enforce
        #
        eval_context = {**JSONModelMeta.__eval_context__,
                        **locals(),
                        **{k: v for frame in reversed(inspect.stack()) for k, v in frame.frame.f_locals.items()},
                        new_cls.__name__: new_cls,
                        **{sc.__name__: sc for sc in ClassHelpers.all_subclasses(JSONModel)}}

        evaluated_anno = {k: mcls._eval(v, eval_context)
                          for k, v in given_anno.items()}
        evaluated_anno = {k: mcls.__annotation_conversions__.get(v, v)
                          for k, v in evaluated_anno.items()}

        def _build_model_elem(k, v) -> ModelElem:
            if isinstance(v, ModelElem):
                return v
            d = defaults.get(k, ())
            _orig, _gener = ClassHelpers.analyze_type(v)
            if d != () and inspect.isclass(_orig) and any(issubclass(_orig, x) for x in (list, dict, set)):
                if len(d) == 0:
                    return ModelElem(v, default_factory=_orig)
                else:
                    return ModelElem(v, default_factory=lambda: _orig(d))
            return ModelElem(v, default=d)

        full_anno: dict[str, ModelElem] = \
            {k: _build_model_elem(k, v)
             for k, v in evaluated_anno.items()}
        clean_anno = {k: v.annotated_type for k, v in full_anno.items()}
        required = [n for n in full_anno.keys() if n not in defaults]
        setattr(new_cls, mcls._FIELDS, tuple(given_anno.keys()))
        setattr(new_cls, mcls._ANNO, given_anno)
        setattr(new_cls, mcls._JSON_MODEL, full_anno)
        setattr(new_cls, "__resolved_anno__", clean_anno)
        setattr(new_cls, "__slots__", required)

        return new_cls

    #

    #

    #


class JSONModel(JSONConvertible, ABC, metaclass=JSONModelMeta):
    _MODEL_CACHE: dict[tuple[Hashable, JSONModelMeta], JSONModelMeta] = dict()

    def __init__(self, **kwargs):
        model = type(self).model()
        ignore_extra = kwargs.pop("_ignore_extra", type(self).__allow_extra_fields__)
        validated = type(self).validate_model(model=model, values=kwargs,
                                              ignore_extra=ignore_extra)
        for k, v in validated.items():
            setattr(self, k, v)

    def __repr__(self) -> str:
        parts = ",".join(f"{k}={getattr(self, k)!r}" for k in self.model().keys())
        return f"{type(self).__name__}({parts})"

    def __str__(self) -> str:
        return repr(self)

    @classmethod
    def model(cls) -> dict[str, ModelElem]:
        """
        Retrieves the JSON model definition for objects of this type.
        """
        return {k: v for k, v in getattr(cls, cls._JSON_MODEL, {}).items()
                if k not in cls.__exclusions__}

    @classmethod
    def from_json(cls, o: JSONObject, **kwargs) -> Self:
        """
        Parses the given JSON into an object of this type (or a subclass)
        """
        model = cls.model()
        copy = {k: v for k, v in o.items()
                if not (mt := model.get(k, None)) or not mt.ephemeral}
        subclass = cls._extract_subclass(copy)
        return subclass(**copy, **kwargs)

    def to_json(self, **kwargs) -> JSONObject:
        """
        Dumps the values in this object into a JSON-friendly dictionary
        """
        model = self.model()
        dumped = {k: elem.dump_value(getattr(self, k), key=k)
                  for k, elem in model.items()
                  if not elem.ignored and not elem.ephemeral
                  and (type(self).__include_defaults_in_json_output__ or
                       not elem.has_default() or getattr(self, k) != elem.default)}

        if not type(self).__allow_null_json_output__:
            dumped = {k: v for k, v in dumped.items()
                      if v is not None}
        if type(self).__include_name_in_json_output__:
            dumped[type(self).__name_field__] = self.model_identity()
        return dumped

    #
    # This section is related to automatically parsing JSONModel subclasses
    # from a JSON object.  These fields are used to identify what subclass
    # it should be parsed as
    #

    @classmethod
    def model_identity(cls) -> JSONType:
        return cls.__name__

    @classmethod
    def _pop_name_from_name_field(cls, o: JSONObject) -> JSONType:
        name_field = cls.__name_field__
        if name_field not in o:
            if cls.__name_field_required__:
                raise JSONModelError(f"Object is missing name field '{name_field}'; cannot "
                                     f"determine class to instantiate.")
            return cls.model_identity()

        return o.pop(name_field)

    @classmethod
    def _subclass_match(cls, value: JSONType, subclass: type[Self]) -> bool:
        return value == subclass.model_identity()

    @classmethod
    def _extract_subclass(cls, o: JSONObject) -> type[Self]:
        name = cls._pop_name_from_name_field(o)
        return cls._extract_subclass_by_name(name)

    @classmethod
    def _extract_subclass_by_name(cls, name: str) -> type[Self]:
        if isinstance(name, Hashable):
            _id = (name, cls)
            if _id in JSONModel._MODEL_CACHE:
                return JSONModel._MODEL_CACHE[_id]

        for subclass in itertools.chain((cls,), ClassHelpers.all_subclasses(cls)):
            if cls._subclass_match(name, subclass):
                if isinstance(name, Hashable):
                    JSONModel._MODEL_CACHE[(name, cls)] = subclass
                return subclass

        raise JSONModelError(f"Unable to find suitable subclass for '{cls.__name__}' matching "
                             f"the name '{name}'; cannot parse JSONModel.")

    #
    # This section is for validating the contents of a model based on its contents.
    # This is generally performed internally, but this is exposed as a public class method
    # to be able to leverage its functionality externally, if needed.
    #

    @classmethod
    def validate_model(cls, model: Mapping[str, ModelElem],
                       values: Mapping[str, Any],
                       ignore_extra: bool = False) -> dict[str, Any]:
        return cls._validate_model(model, values, ignore_extra)

    @classmethod
    def _validate_model(cls, model: Mapping[str, ModelElem],
                        values: Mapping[str, Any],
                        ignore_extra: bool) -> dict[str, Any]:
        result: dict[str, Any] = dict()
        values = {**values}
        for k, m in model.items():
            if m.ignored:
                values.pop(k, None)
                continue
            if k not in values:
                if not m.has_default():
                    raise JSONModelError(f"Missing required key '{k}' on '{cls.__name__}'.")
                val = m.default
            else:
                val = values.pop(k)
            try:
                val = m.validate(val, key=k)
            except ModelElemError as e:
                raise JSONModelError(f"Model error on key '{k}' of '{cls.__name__}': {str(e)}")

            result[k] = val

        if not ignore_extra and any(k for k in values.keys() if k not in cls.__exclusions__):
            raise JSONModelError(f"The following keys are not found in the model for '{cls.__name__}': "
                                 f"{','.join(values.keys())}")

        return result



from __future__ import annotations

import functools
import io
from typing import TextIO
from SprelfJSON.JSONDefinitions import JSONType, JSONable

YAMLError: type[Exception] | None = None


@functools.cache
def yaml():
    global YAMLError
    try:
        from ruamel.yaml import YAML, YAMLError as ye
        YAMLError = ye
        y = YAML(typ="safe")
        y.indent = 2
        y.encoding = "utf-8"
        y.default_flow_style = False
        return y
    except ImportError:
        raise ImportError("ruamel.yaml is required for YAML support.  Try install sprelf-json[yaml].")


def dumps(obj: JSONType | JSONable) -> str:
    with io.StringIO() as f:
        yaml().dump(jsonify(obj), f)
        return f.getvalue()


def dump(f: TextIO, obj: JSONType | JSONable) -> None:
    yaml().dump(jsonify(obj), f)


def load(s: str | TextIO, default: JSONType, raise_errors: bool = False) -> JSONType:
    try:
        return yaml().load(s)
    except YAMLError:
        if raise_errors:
            raise
        return default if default is not None else dict()


def loads(s: str | TextIO, default: JSONType, raise_errors: bool = False) -> JSONType:
    return load(s, default, raise_errors)


def jsonify(obj: JSONType | JSONable) -> JSONType:
    return obj.to_json() if isinstance(obj, JSONable) else obj



from __future__ import annotations

import typing_inspect
from typing import Iterable, TypeAlias, get_origin, get_args, Any, TypeVar, Iterator, Sequence, Collection
from types import ModuleType
import collections
import typing
import types
import inspect
import pydoc
import builtins
import sys

from SprelfJSON.Helpers.Decorators import yield_unique


T = TypeVar('T')


def analyze_type(t: type) -> tuple[type, tuple[type, ...]]:
    o = get_origin(t) or t
    a = get_args(t)
    return o, a


def check_subclass(check: Any, expected: Any) -> bool:
    origin, args = analyze_type(expected)
    return check_generic_subclass(check, origin, *args)


def check_generic_subclass(check: Any, expected_origin: Any, *expected_args: Any) -> bool:
    if not isinstance(check, type):
        return False

    if expected_origin == Any:
        return True

    origin, args = analyze_type(check)

    if typing_inspect.is_union_type(expected_origin):
        return any(check_subclass(check, a) for a in expected_args)

    if not issubclass(origin, expected_origin):
        return False

    if len(expected_args) != len(args):
        return False

    if len(expected_args) == 0:
        return True

    return all(check_subclass(c, e) for c, e in zip(args, expected_args))


def check_instance(check: Any, expected: Any) -> bool:
    origin, args = analyze_type(expected)
    return check_generic_instance(check, origin, *args)


def check_generic_instance(check: Any, expected_origin: Any, *expected_args: Any) -> bool:
    if typing_inspect.is_union_type(expected_origin):
        return any(check_instance(check, a) for a in expected_args)

    if expected_origin == Any:
        return True

    if not isinstance(check, expected_origin):
        return False

    if len(expected_args) == 0:
        return True

    if expected_origin in (Iterable, Iterator, Sequence, Collection):
        return all(check_instance(x, expected_args[0]) for x in check)

    if issubclass(expected_origin, type):
        return inspect.isclass(check) and check_subclass(check, expected_args[0])

    if issubclass(expected_origin, tuple) and isinstance(check, collections.abc.Collection):
        if len(expected_args) == 2 and expected_args[1] is Ellipsis:
            return all(check_instance(x, expected_args[0]) for x in check)
        elif len(expected_args) != len(check):
            return False
        return all(check_instance(x, a) for x, a in zip(check, expected_args))

    if issubclass(expected_origin, Iterable) and isinstance(check, Iterable):
        return all(check_instance(x, expected_args[0]) for x in check)

    elif issubclass(expected_origin, collections.abc.Mapping) and isinstance(check, collections.abc.Mapping):
        key_type, val_type = expected_args
        return all(check_instance(k, key_type) and check_instance(v, val_type)
                   for k, v in check.items())

    return False


#


def as_generic(origin: Any, *generics: Any) -> Any:
    if len(generics) == 0:
        return origin
    try:
        return origin[generics]
    except TypeError:
        if origin == types.UnionType:
            return typing.Union[generics]

        # fallback to typing equivalents if needed
        origin_names = (origin.__name__, origin.__name__.capitalize())
        for origin_name in origin_names:
            for module in (typing, types):
                typing_origin = getattr(module, origin_name, None)
                if typing_origin:
                    return typing_origin[generics]
        raise

#


def type_as_string(t: type) -> str:
    origin, args = analyze_type(t)

    if origin is None:
        return t.__name__

    base_name = origin.__name__
    if len(args) == 0:
        return base_name

    arg_str = ", ".join(type_as_string(a) for a in args)
    return f"{base_name}[{arg_str}]"


#


@yield_unique
def all_subclasses(t: type) -> Iterable[type]:
    for c in t.__subclasses__():
        yield c
        yield from all_subclasses(c)


def get_module(t: type | ModuleType) -> ModuleType:
    return inspect.getmodule(t)


def full_name(t: type) -> str:
    return get_module(t).__name__ + "." + t.__name__


def locate_class(name: str) -> type | None:
    # Try pydoc.locate first (handles fully qualified names)
    obj = pydoc.locate(name)
    if isinstance(obj, type):
        return obj

    # Split the name (in case it's like "module.Class")
    split = name.split(".")
    if len(split) > 1:
        module_path = ".".join(split[:-1])
        class_name = split[-1]

        module = pydoc.locate(module_path)
        if isinstance(module, ModuleType):
            return getattr(module, class_name, None)

    # Try builtins
    if hasattr(builtins, name):
        obj = getattr(builtins, name)
        if isinstance(obj, type):
            return obj

    # Search through all currently loaded modules
    for mod in sys.modules.values():
        if mod and isinstance(mod, ModuleType) and hasattr(mod, name):
            obj = getattr(mod, name)
            if isinstance(obj, type):
                return obj

    # Optionally: scan local and global scopes in the call stack
    for frame_info in inspect.stack():
        frame = frame_info.frame
        if name in frame.f_locals:
            obj = frame.f_locals[name]
            if isinstance(obj, type):
                return obj
        if name in frame.f_globals:
            obj = frame.f_globals[name]
            if isinstance(obj, type):
                return obj

    # Not found
    return None



from SprelfJSON.JSONModel.JSONModelError import JSONModelError
from SprelfJSON.JSONModel.ModelElem import ModelElemError, ModelElem, AlternateModelElem
from SprelfJSON.JSONModel.JSONModel import JSONModel



from SprelfJSON.JSONDefinitions import JSONType, JSONObject, JSONValue, JSONArray, JSONContainer, FieldPath, \
    SprelfJSONError, is_json_type, JSONable, JSONConvertible, JSONObjectLike, JSONArrayLike, JSONValueLike, \
    JSONContainerLike, JSONLike
from SprelfJSON.JSONUtils import get as json_get
from SprelfJSON.JSONModel import *
from SprelfJSON.Helpers import *
from SprelfJSON.Objects import *



from __future__ import annotations

from typing import Generic, TypeVar, Any, Iterator, Iterable, Self
import copy
import functools

T = TypeVar('T')


#


class Ephemeral(Generic[T]):
    """
    This class is intended to be a wrapper over any kind of object, except that
    it will not be parsed or dumped as part of a JSONModel; it will only ever
    live in memory and then vanish.  As a result, the contents wrapped by this
    do not need to be JSON-serializable, as they will never interact with
    JSON, by design.

    It is intended that this wrapper mimics the functionality and identify
    of whatever it wraps, except for passing isinstance checks;
    for that, see Ephemeral.is_ephemeral()
    """
    __ephm_field__: str = "_ephm_value"
    __slots__ = (__ephm_field__,)
    __is_ephemeral__ = True

    def __new__(cls, value: T):
        return super().__new__(_build(value))

    def __init__(self, value: T):
        object.__setattr__(self, type(self).__ephm_field__, value)

    @property
    def value(self) -> T:
        return object.__getattribute__(self, type(self).__ephm_field__)

    @property
    def V(self) -> T:
        return self.value

    # Attribute delegation
    def __getattr__(self, name: str) -> Any:
        return getattr(self.value, name)

    def __setattr__(self, name: str, value: Any) -> None:
        if name == type(self).__ephm_field__:
            object.__setattr__(self, name, value)
        else:
            setattr(self.value, name, value)

    def __delattr__(self, name: str) -> None:
        if name == type(self).__ephm_field__:
            raise AttributeError("Cannot delete Ephemeral internal state")
        delattr(self.value, name)

    def __dir__(self) -> list[str]:
        own = set(super().__dir__())
        try:
            other = set(dir(self.value))
        except Exception:
            other = set()
        return sorted(own | other)

    def __copy__(self) -> Ephemeral[T]:
        try:
            v = copy.copy(self.value)
        except Exception:
            v = self.value
        return type(self)(v)

    def __deepcopy__(self, memo) -> Ephemeral[T]:
        try:
            v = copy.deepcopy(self.value, memo)
        except Exception:
            v = self.value
        return type(self)(v)



    @classmethod
    def is_ephemeral(cls, obj: Any) -> bool:
        return isinstance(obj, cls) or getattr(obj, "__is_ephemeral__", False)

    @classmethod
    def unwrap(cls, o: Ephemeral[T] | T) -> T:
        return o.value if isinstance(o, Ephemeral) else o


def _build(value: T) -> type[Ephemeral[T]]:
    key = sorted((k
                  for k in ("__len__", "__iter__", "__contains__", "__enter__", "__exit__",
                            "__index__", "__int__", "__float__", "__complex__", "__bytes__",
                            "__hash__")
                  if hasattr(value, k)))
    new_cls: type[Ephemeral] = _get_proxy(tuple(key))
    return new_cls


@functools.cache
def _get_proxy(key: tuple[str, ...]) -> type[Ephemeral]:
    namespace: dict[str, Any] = {}

    namespace["__repr__"] = lambda self: f"{type(self).__name__}({self.value!r})"
    namespace["__str__"] = lambda self: f"{str(self.value)}"
    namespace["__bool__"] = lambda self: bool(self.value)

    namespace["__eq__"] = lambda self, other: self.value == (other.value if isinstance(other, Ephemeral) else other)
    namespace["__ne__"] = lambda self, other: self.value != (other.value if isinstance(other, Ephemeral) else other)
    namespace["__lt__"] = lambda self, other: self.value < (other.value if isinstance(other, Ephemeral) else other)
    namespace["__le__"] = lambda self, other: self.value <= (other.value if isinstance(other, Ephemeral) else other)
    namespace["__gt__"] = lambda self, other: self.value > (other.value if isinstance(other, Ephemeral) else other)
    namespace["__ge__"] = lambda self, other: self.value >= (other.value if isinstance(other, Ephemeral) else other)

    for dund in key:

        # namespace[dund] = lambda self: getattr(self.value, dund)()
        match dund:
            case "__len__":
                def __len__(self):
                    return len(self.value)

                namespace[dund] = __len__
            case "__iter__":
                def __iter__(self) -> Iterator:
                    return iter(self.value)

                namespace[dund] = __iter__
            case "__contains__":
                def __contains__(self, item: Any) -> bool:
                    return item in self.value

                namespace[dund] = __contains__
            case "__call__":
                def __call__(self, *args: Any, **kwargs: Any) -> Any:
                    return self.value(*args, **kwargs)

                namespace[dund] = __call__
            case "__enter__":
                def __enter__(self) -> Ephemeral[T]:
                    return self.value.__enter__()

                namespace[dund] = __enter__
            case "__exit__":
                def __exit__(self, *args: Any) -> Ephemeral[T]:
                    return self.value.__exit__(*args)

                namespace[dund] = __exit__
            case "__index__":
                def __index__(self) -> int:
                    return self.value.__index__()

                namespace[dund] = __index__
            case "__int__":
                def __int__(self) -> int:
                    return int(self.value)

                namespace[dund] = __int__
            case "__float__":
                def __float__(self) -> float:
                    return float(self.value)

                namespace[dund] = __float__
            case "__complex__":
                def __complex__(self) -> complex:
                    return complex(self.value)

                namespace[dund] = __complex__
            case "__bytes__":
                def __bytes__(self) -> bytes:
                    return bytes(self.value)

                namespace[dund] = __bytes__
            case "__hash__":
                def __hash__(self) -> int:
                    return hash(self.value)

                namespace[dund] = __hash__

    new_cls: type[Ephemeral] = type(f"EphmProxy[{','.join(key)}]", (Ephemeral,), namespace)
    return new_cls


from SprelfJSON.JSONDefinitions import JSONObject, JSONValue, JSONType, JSONArray, JSONContainer, FieldPath


def get(o: JSONValue, path: FieldPath, default: JSONType = None):
    """
    Retrieves the value at the specified field path in the given JSON object, if it exists.  If not
    found, will return the given default value instead.  Components of the field path can either be
    strings (to get keys from an object) or integers (to get elements from an array).  Field path
    may be given either as a tuple of fields to navigate, or a single string where each field
    is separated by a (.) period.

    :param o: The JSON value to extract the value from.
    :param path: The path at which to extract the value.
    :param default: Optional.  The default value to return if unsuccessful in finding the specified field.
    Defaults to None.
    :return: The extract value, or the default value if not successful.
    """
    if isinstance(path, str):
        path = tuple(path.split("."))
    if len(path) == 0:
        return o
    if isinstance(o, JSONValue):
        return default
    curr, *rest = path
    if isinstance(o, list):
        if not isinstance(curr, int):
            try:
                curr = int(curr)
            except ValueError:
                return default
        if not (0 <= curr < len(o)):
            return default
        return get(o[curr], rest, default)
    elif isinstance(o, dict):
        if not isinstance(curr, str):
            return default
        if curr not in o:
            return default
        return get(o[curr], rest, default)
